from datetime import datetime
from typing import List, Optional, Set, Tuple, Union

from viur.datastore.transport import Get, Put, RunInTransaction

from viur.datastore.types import Entity, Key, currentDbAccessLog, currentTransaction


def fixUnindexableProperties(entry: Entity) -> Entity:
    """
        Recursively walk the given Entity and add all properties to the list of unindexed properties if they contain
        a string longer than 500 bytes (which is maximum size of a string that can be indexed). The datastore would
        return an error otherwise.
    :param entry: The entity to fix (inplace)
    :return: The fixed entity
    """

    def hasUnindexableProperty(prop):
        if isinstance(prop, dict):
            return any([hasUnindexableProperty(x) for x in prop.values()])
        elif isinstance(prop, list):
            return any([hasUnindexableProperty(x) for x in prop])
        elif isinstance(prop, (str, bytes)):
            return len(prop) >= 500
        else:
            return False

    resList = set()
    for k, v in entry.items():
        if hasUnindexableProperty(v):
            if isinstance(v, dict):
                innerEntry = Entity()
                innerEntry.update(v)
                entry[k] = fixUnindexableProperties(innerEntry)
                if isinstance(v, Entity):
                    innerEntry.key = v.key
            else:
                resList.add(k)
    entry.exclude_from_indexes = resList
    return entry


def normalizeKey(key: Union[None, Key]) -> Union[None, Key]:
    """
        Normalizes a datastore key (replacing _application with the current one)

        :param key: Key to be normalized.
        :return: Normalized key in string representation.
    """
    if key is None:
        return None
    if key.parent:
        parent = normalizeKey(key.parent)
    else:
        parent = None
    return Key(key.kind, key.id_or_name, parent=parent)


def keyHelper(inKey: Union[Key, str, int], targetKind: str,
              additionalAllowedKinds: Union[List[str], Tuple[str]] = (),
              adjust_kind: bool = False) -> Key:
    if isinstance(inKey, Key):
        if inKey.kind != targetKind and inKey.kind not in additionalAllowedKinds:
            if not adjust_kind:
                raise ValueError(f"Kind mismatch: {inKey.kind!r} != {targetKind!r} (or in {additionalAllowedKinds!r})")
            inKey.kind = targetKind
        return inKey
    elif isinstance(inKey, str):
        # Try to parse key from str
        try:
            decodedKey = normalizeKey(Key.from_legacy_urlsafe(inKey))
        except:
            decodedKey = None

        # If it did decode, recall keyHelper with Key object
        if decodedKey:
            return keyHelper(
                decodedKey,
                targetKind=targetKind,
                additionalAllowedKinds=additionalAllowedKinds,
                adjust_kind=adjust_kind
            )

        # otherwise, construct key from str or int
        if inKey.isdigit():
            inKey = int(inKey)

        return Key(targetKind, inKey)
    elif isinstance(inKey, int):
        return Key(targetKind, inKey)

    raise NotImplementedError(f"Unsupported key type {type(inKey)}")


def IsInTransaction() -> bool:
    return currentTransaction.get() is not None


def GetOrInsert(key: Key, **kwargs) -> Entity:
    """
        Either creates a new entity with the given key, or returns the existing one.

        Its guaranteed that there is no race-condition here; it will never overwrite an
        previously created entity. Extra keyword arguments passed to this function will be
        used to populate the entity if it has to be created; otherwise they are ignored.

        :param key: The key which will be fetched or created.
        :returns: Returns the fetched or newly created Entity.
    """

    def txn(key, kwargs):
        obj = Get(key)
        if not obj:
            obj = Entity(key)
            for k, v in kwargs.items():
                obj[k] = v
            Put(obj)
        return obj

    if IsInTransaction():
        return txn(key, kwargs)
    return RunInTransaction(txn, key, kwargs)


def encodeKey(key: Key) -> str:
    """
        Return the given key encoded as string (mimicking the old str() behaviour of keys)
    """
    # todo: Should we make a deprecation warning when this function is used?
    return str(key)


def acquireTransactionSuccessMarker() -> str:
    """
        Generates a token that will be written to the firestore (under "viur-transactionmarker") if the transaction
        completes successfully. Currently only used by deferredTasks to check if the task should actually execute
        or if the transaction it was created in failed.
        :return: Name of the entry in viur-transactionmarker
    """
    txn = currentTransaction.get()
    assert txn, "acquireTransactionSuccessMarker cannot be called outside an transaction"
    marker = txn["key"]  # binascii.b2a_hex(txn["key"]).decode("ASCII")
    if not "viurTxnMarkerSet" in txn:
        e = Entity(Key("viur-transactionmarker", marker))
        e["creationdate"] = datetime.utcnow()
        Put(e)
        txn["viurTxnMarkerSet"] = True
    return marker


def startDataAccessLog() -> Set[Union[Key, str]]:
    """
        Clears our internal access log (which keeps track of which entries have been accessed in the current
        request). The old set of accessed entries is returned so that it can be restored with
        :func:`server.db.popAccessData` in case of nested caching. You must call popAccessData afterwards, otherwise
        we'll continue to log all entries accessed in subsequent request on the same thread!
        :return: Set of old accessed entries
    """
    old = currentDbAccessLog.get(set())
    currentDbAccessLog.set(set())
    return old


def endDataAccessLog(outerAccessLog: Optional[Set[Union[Key, str]]] = None) -> Optional[Set[Union[Key, str]]]:
    """
        Retrieves the set of entries accessed so far. To clean up and restart the log, call :func:`viur.datastore.startAccessDataLog`.
        If you called :func:`server.db.startAccessDataLog` before, you can re-apply the old log using
        the outerAccessLog param. Otherwise, it will disable the access log.
        :param outerAccessLog: State of your log returned by :func:`server.db.startAccessDataLog`
        :return: Set of entries accessed
    """
    res = currentDbAccessLog.get()
    if isinstance(outerAccessLog, set):
        currentDbAccessLog.set((outerAccessLog or set()).union(res))
    else:
        currentDbAccessLog.set(None)
    return res



from viur.datastore.config import conf as config
from viur.datastore.errors import *
from viur.datastore import cache
from viur.datastore.query import Query
from viur.datastore.transport import AllocateIDs, Delete, Get, Put, RunInTransaction, Count
from viur.datastore.types import (
    currentDbAccessLog,
    DATASTORE_BASE_TYPES,
    Entity,
    KEY_SPECIAL_PROPERTY,
    Key,
    SortOrder,
    SkelListRef,
    QueryDefinition)
from viur.datastore.utils import (
    fixUnindexableProperties,
    normalizeKey,
    keyHelper,
    IsInTransaction,
    GetOrInsert,
    encodeKey,
    acquireTransactionSuccessMarker,
    startDataAccessLog,
    endDataAccessLog)

import logging

# silencing requests' debugging
logging.getLogger("requests").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)

__all__ = [
    "KEY_SPECIAL_PROPERTY",
    "DATASTORE_BASE_TYPES",
    "SortOrder",
    "SkelListRef",
    "Entity",
    "QueryDefinition",
    "Key",
    "Query",
    "fixUnindexableProperties",
    "normalizeKey",
    "keyHelper",
    "Get",
    "Count",
    "Put",
    "Delete",
    "RunInTransaction",
    "IsInTransaction",
    "currentDbAccessLog",
    "GetOrInsert",
    "encodeKey",
    "acquireTransactionSuccessMarker",
    "AllocateIDs",
    "config",
    "startDataAccessLog",
    "endDataAccessLog",
    "ViurDatastoreError",
    "AbortedError",
    "CollisionError",
    "DeadlineExceededError",
    "FailedPreconditionError",
    "InternalError",
    "InvalidArgumentError",
    "NotFoundError",
    "PermissionDeniedError",
    "ResourceExhaustedError",
    "UnauthenticatedError",
    "UnavailableError",
    "NoMutationResultsError",
    "is_viur_datastore_request_ok",
    "cache",
]



"""
The constants, global variables and container classes used in the datastore api
"""
from __future__ import annotations

import typing as t
from base64 import urlsafe_b64decode, urlsafe_b64encode
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum
from typing import Dict, List, Optional, Set, Tuple, Union

import google.auth
from google.cloud.datastore import _app_engine_key_pb2

if t.TYPE_CHECKING:
    from viur.core.skeleton import SkeletonInstance

# The property name pointing to an entities key in a query
KEY_SPECIAL_PROPERTY = "__key__"
# List of types that can be used in a datastore query
DATASTORE_BASE_TYPES = Union[None, str, int, float, bool, datetime, date, time, 'Key']  #
# Pointer to the current transaction this thread may be currently in
currentTransaction = ContextVar("CurrentTransaction", default=None)
# If set to a set for the current thread/request, we'll log all entities / kinds accessed
currentDbAccessLog: ContextVar[Optional[Set[Union[Key, str]]]] = ContextVar("Database-Accesslog", default=None)
# The current projectID, which can't be imported from transport.pyx
_, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])


class SortOrder(Enum):
    Ascending = 1  # Sort A->Z
    Descending = 2  # Sort Z->A
    InvertedAscending = 3  # Fetch Z->A, then flip the results (useful in pagination to go from a start cursor backwards)
    InvertedDescending = 4  # Fetch A->Z, then flip the results (useful in pagination)


class SkelListRef(list):
    """
        This class is used to hold multiple skeletons together with other, commonly used information.

        SkelLists are returned by Skel().all()...fetch()-constructs and provide additional information
        about the data base query, for fetching additional entries.

        :ivar cursor: Holds the cursor within a query.
        :vartype cursor: str
    """

    __slots__ = ["baseSkel", "getCursor", "get_orders", "customQueryInfo", "renderPreparation"]

    def __init__(self, baseSkel: t.Optional["SkeletonInstance"] = None):
        """
            :param baseSkel: The baseclass for all entries in this list
        """
        super(SkelListRef, self).__init__()
        self.baseSkel = baseSkel or {}
        self.getCursor = lambda: None
        self.get_orders = lambda: None
        self.renderPreparation = None
        self.customQueryInfo = {}


class Key:
    """
        The python representation of one datastore key. Unlike the original implementation, we don't store a
        reference to the project the key lives in. This is always expected to be the current project as ViUR
        does not support accessing data in multiple projects.
    """

    __slots__ = ["id", "name", "kind", "parent"]

    def __init__(self, kind: str, subKey: Union[int, str, None] = None, parent: 'Key' = None):
        super().__init__()
        self.kind = kind
        self.id = None
        self.name = None
        if isinstance(subKey, int):
            self.id = subKey
        elif isinstance(subKey, str):
            if subKey.isdigit():
                self.id = int(subKey)
            else:
                self.name = subKey
        elif subKey is not None:
            raise ValueError(f"Invalid argument type {subKey = }")
        self.parent = parent

    @property
    def id_or_name(self) -> Union[None, str, int]:
        """
            :return: This key's id or name (or none, if this key is partial)
        """
        return self.id or self.name

    def __str__(self):
        return self.to_legacy_urlsafe().decode("ASCII")

    def __repr__(self):
        return "<viur.datastore.Key %s/%s, parent=%s>" % (self.kind, self.id_or_name, self.parent)

    def __hash__(self):
        return hash("%s.%s.%s" % (self.kind, self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, Key) and self.kind == other.kind and self.id == other.id and self.name == other.name \
            and self.parent == other.parent

    def to_legacy_urlsafe(self) -> bytes:
        """
            Converts this key into the (urlsafe) protobuf string representation.
            :return: The urlsafe string representation of this key
        """
        currentKey = self
        pathElements = []
        while currentKey:
            pathElements.insert(0, _app_engine_key_pb2.Path.Element(
                type=currentKey.kind,
                id=currentKey.id,
                name=currentKey.name,
            ))
            currentKey = currentKey.parent
        reference = _app_engine_key_pb2.Reference(
            app=projectID,
            path=_app_engine_key_pb2.Path(element=pathElements),
        )
        raw_bytes = reference.SerializeToString()
        return urlsafe_b64encode(raw_bytes).strip(b"=")

    @property
    def is_partial(self) -> bool:
        """
            Checks if this key is partial (ie it belongs to an entity that has not been saved to the datastore).
            If the entity is saved, this key will be replaced by a full key (having an id or a name assigned)
            :return: True if this key is partial
        """
        return self.id_or_name is None

    @classmethod
    def from_legacy_urlsafe(cls, strKey: str) -> Key:
        """
            Parses the string representation generated by :meth:to_legacy_urlsafe into a new Key object
            :param strKey: The string key to parse
            :return: The new Key object constructed from the string key
        """
        urlsafe = strKey.encode("ASCII")
        padding = b"=" * (-len(urlsafe) % 4)
        urlsafe += padding
        raw_bytes = urlsafe_b64decode(urlsafe)
        reference = _app_engine_key_pb2.Reference()
        reference.ParseFromString(raw_bytes)
        resultKey = None
        for elem in reference.path.element:
            resultKey = Key(elem.type, elem.id or elem.name, parent=resultKey)
        return resultKey


class Entity(dict):
    """
        The python representation of one datastore entity. The values of this entity are stored inside this dictionary,
        while the meta-data (it's key, the list of properties excluded from indexing and our version) as property values.
    """
    __slots__ = ["key", "_exclude_from_indexes", "version"]

    def __init__(self, key: Optional[Key] = None, exclude_from_indexes: Optional[Set[str]] = None):
        super(Entity, self).__init__()
        assert not key or isinstance(key, Key), "Key must be a Key-Object (or None for an embedded entity)"
        self.key = key
        self.exclude_from_indexes = exclude_from_indexes or set()
        assert isinstance(self.exclude_from_indexes, set)
        self.version = None

    @property
    def exclude_from_indexes(self) -> set[str]:
        return self._exclude_from_indexes

    @exclude_from_indexes.setter
    def exclude_from_indexes(self, value: set[str] | list[str] | tuple[str]) -> None:
        self._exclude_from_indexes = set(value)


@dataclass
class QueryDefinition:
    """
        A single Query that will be run against the datastore.
    """
    kind: Optional[str]  # The datastore kind to run the query on. Can be None for kindles queries.
    filters: Dict[str, DATASTORE_BASE_TYPES]  # A dictionary of constrains to apply to the query.
    orders: List[Tuple[str, SortOrder]]  # The list of fields to sort the results by.
    distinct: Union[None, List[str]] = None  # If set, a list of fields that we should return distinct values of
    limit: int = 30  # The maximum amount of entities that should be returned
    startCursor: Optional[str] = None  # If set, we'll only return entities that appear after this cursor in the index.
    endCursor: Optional[str] = None  # If set, we'll only return entities up to this cursor in the index.
    currentCursor: Optional[
        str] = None  # Will be set after this query has been run, pointing after the last entity returned



# distutils: sources = src/viur/datastore/simdjson.cpp
# distutils: language = c++
# cython: language_level=3
from copy import copy
from viur.datastore import cache
import google.auth
import requests
from libcpp cimport bool as boolean_type
from viur.datastore.types import currentTransaction, Entity, Key, QueryDefinition, currentDbAccessLog
from viur.datastore.config import conf
from viur.datastore.errors import *
from cython.operator cimport preincrement, dereference
from libc.stdint cimport int64_t, uint64_t
from datetime import datetime, timezone
from cpython.bytes cimport PyBytes_AsStringAndSize
import json
from base64 import b64decode, b64encode
from typing import Union, List, Any
from requests.exceptions import ConnectionError as RequestsConnectionError
import logging
from time import sleep
## Start of CPP-Imports required for the simdjson->python bridge

cdef extern from "Python.h":
    object PyUnicode_FromStringAndSize(const char *u, Py_ssize_t size)

cdef extern from "string_view" namespace "std":
    cppclass stringView "std::string_view":
        char * data()
        int length()
        int compare(char *)

cdef extern from "simdjson.h" namespace "simdjson::error_code":
    cdef enum simdjsonErrorCode "simdjson::error_code":
        SUCCESS,
        NO_SUCH_FIELD

cdef extern from "simdjson.h" namespace "simdjson::dom::element_type":
    cdef enum simdjsonElementType "simdjson::dom::element_type":
        OBJECT,
        ARRAY,
        STRING,
        INT64,
        UINT64,
        DOUBLE,
        BOOL,
        NULL_VALUE

cdef extern from "simdjson.h" namespace "simdjson::simdjson_result":
    cdef cppclass simdjsonResult "simdjson::simdjson_result<simdjson::dom::element>":
        simdjsonErrorCode error()
        simdjsonElement value()

cdef extern from "simdjson.h" namespace "simdjson::dom":
    cdef cppclass simdjsonObject "simdjson::dom::object":
        cppclass iterator:
            iterator()
            iterator operator++()
            bint operator !=(iterator)
            stringView key()
            simdjsonElement value()
        simdjsonObject()
        iterator begin()
        iterator end()

    cdef cppclass simdjsonArray "simdjson::dom::array":
        cppclass iterator:
            iterator()
            iterator operator++()
            bint operator !=(iterator)
            simdjsonElement operator *()
        simdjsonArray()
        iterator begin()
        iterator end()
        simdjsonElement at(int) except +
        simdjsonElement at_pointer(const char *) except +
        int size()

    cdef cppclass simdjsonElement "simdjson::dom::element":
        simdjsonElement()
        simdjsonElementType type() except +
        boolean_type get_bool() except +
        int64_t get_int64() except +
        uint64_t get_uint64() except +
        double get_double() except +
        stringView get_string() except +
        simdjsonArray get_array() except +
        simdjsonObject get_object() except +
        simdjsonElement at_key(const char *) except +  # Raises if key not found
        simdjsonResult at_pointer(const char *)  # Same as at_key - sets result error code if key not found

    cdef cppclass simdjsonParser "simdjson::dom::parser":
        simdjsonParser()
        simdjsonElement parse(const char * buf, size_t len, boolean_type realloc_if_needed) except +

## End of C-Imports


credentials, projectID = google.auth.default(scopes=["https://www.googleapis.com/auth/datastore"])
_http_internal = google.auth.transport.requests.AuthorizedSession(
    credentials,
    refresh_timeout=300,
)

def authenticated_request(url: str, data: bytes) -> requests.Response:
    """
        Runs one http request to the datastore rest api, authenticated with the current projects service account.
        Will retry up to three times in case the connection cannot be established the first time.

        :param url: The url to post the data to
        :param data:: The data to include in the post request
        :return: The Response object
    """
    for i in range(0, 3):
        try:
            return _http_internal.post(
                url=url,
                data=data,
            )
        except RequestsConnectionError:
            logging.debug("Retrying http post request to datastore")
            if i == 2:
                raise
            continue

def keyToPath(key: Key) -> List[dict]:
    """
        Converts a Key object to the PathElements expected by the rest API.
        See https://cloud.google.com/datastore/docs/reference/data/rest/v1/Key#PathElement

        :param key: The key object to convert
        :return: The list of path elements corresponding to this key
    """
    res = []
    while key:
        res.insert(0,
                   {
                       "kind": key.kind,
                       "id": key.id,
                   } if key.id else {
                       "kind": key.kind,
                       "name": key.name
                   } if key.name else {
                       "kind": key.kind,
                   }
                   )
        key = key.parent
    return res

def pythonPropToJson(v) -> dict:
    """
        Convert a python type to the object representation expected by the datastore rest API.
        See https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery#Value

        :param v: The python object to convert
        :return: The representation of that object as expected by the rest api.
    """
    if v is True or v is False:
        return {
            "booleanValue": v
        }
    elif v is None:
        return {"nullValue": None}
    elif isinstance(v, int):
        return {
            "integerValue": str(v)
        }
    elif isinstance(v, float):
        return {
            "doubleValue": v
        }
    elif isinstance(v, str):
        return {
            "stringValue": v
        }
    elif isinstance(v, Key):
        return {
            "keyValue": {
                "partitionId": {
                    "project_id": projectID,
                },
                "path": keyToPath(v)
            }
        }
    elif isinstance(v, datetime):
        return {
            "timestampValue": v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        }
    elif isinstance(v, Entity):
        excludedProperties: set = set(v.exclude_from_indexes)
        properties = {}
        for dictKey, dictValue in v.items():
            if isinstance(dictValue, list) and dictKey in excludedProperties:
                # Lists cannot be not indexed (but each of its value can be). So we have to forward our not-indexed
                # flag to our children. See https://github.com/googleapis/google-cloud-node/issues/2615.
                resList = []
                for value in dictValue:
                    resDict = pythonPropToJson(value)
                    resDict["excludeFromIndexes"] = True
                    resList.append(resDict)
                properties[dictKey] = {
                    "arrayValue": {
                        "values": resList
                    }
                }
            else:
                resDict = pythonPropToJson(dictValue)
                if dictKey in excludedProperties:
                    resDict["excludeFromIndexes"] = True
                properties[dictKey] = resDict
        return {
            "entityValue": {
                "key": pythonPropToJson(v.key)["keyValue"] if isinstance(v, Entity) and v.key else None,
                "properties": properties
            }
        }
    elif isinstance(v, dict):
        # We hande dicts separately as they don't have keys nor can keys be excluded from being indexed
        return {
            "entityValue": {
                "key": None,
                "properties": {
                    dictKey: pythonPropToJson(dictValue) for dictKey, dictValue in v.items()
                }
            }
        }
    elif isinstance(v, list):
        return {
            "arrayValue": {
                "values": [pythonPropToJson(x) for x in v]
            }
        }
    elif isinstance(v, bytes):
        return {
            "blobValue": b64encode(v).decode("ASCII")
        }
    assert False, "%s (%s) is not supported" % (v, type(v))

cdef inline object toPyStr(stringView strView):
    """
        Converts a cpp stringview to a python str object
        :param strView: The stringview object
        :return: The corresponding python string object
    """
    return PyUnicode_FromStringAndSize(
        strView.data(),
        strView.length()
    )

cdef inline object parseKey(simdjsonElement v):
    """
        Parses a simdJsonObject representing a key to a datastore.Key instance.

        :param v: The simdJsonElement containing the key
        :return: The corresponding datastore.Key instance
    """
    cdef simdjsonArray arr
    cdef simdjsonArray.iterator arrayIt, arrayItEnd
    pathArgs = []
    arr = v.at_key("path").get_array()
    arrayIt = arr.begin()
    arrayItEnd = arr.end()
    while arrayIt != arrayItEnd:
        element = dereference(arrayIt)
        if element.at_pointer("/id").error() == SUCCESS:
            pathArgs.append(
                (toPyStr(element.at_key("kind").get_string()), int(toPyStr(element.at_key("id").get_string())))
            )
        elif element.at_pointer("/name").error() == SUCCESS:
            pathArgs.append(
                (toPyStr(element.at_key("kind").get_string()), toPyStr(element.at_key("name").get_string()))
            )
        else:
            # We read an incomplete/null key from the datastore. This is likely a bug.
            # Can happen if you manually created a key like datastore.Key("kind", 0).
            logging.error("We read an incomplete/null key from the datastore. This is likely a bug!")
            pathArgs.append((toPyStr(element.at_key("kind").get_string()), 0))
        preincrement(arrayIt)
    key = None
    for pathElem in pathArgs:
        key = Key(*pathElem, parent=key)
    return key

cdef inline object toEntityStructure(simdjsonElement v, boolean_type isInitial = False):
    """
        Parses a simdJsonElement into the corresponding python datatypes.
        :param v: The simdJsonElement to parse
        :param isInitial: If true, we'll return a dictionary of key->entity instead of a list of entities
        :return: The corresponding python datatype.

        TODO: While working for now, this should probably refactored to return the parsed value and it's index flag
            to avoid additional loops over objects.
    """
    cdef simdjsonElementType et = v.type()
    cdef simdjsonArray.iterator arrayIt, arrayItEnd
    cdef simdjsonObject.iterator objIterStart, objIterStartInner, objIterEnd, objIterEndInner
    cdef simdjsonElement element
    cdef simdjsonObject outerObject, innerObject
    cdef stringView strView, bytesView
    cdef simdjsonResult tmpResult
    if et == OBJECT:
        outerObject = v.get_object()
        objIterStart = outerObject.begin()
        objIterEnd = outerObject.end()
        while objIterStart != objIterEnd:
            strView = objIterStart.key()
            if strView.compare("entity") == 0 or strView.compare("entityValue") == 0:
                e = Entity()
                excludeList = set()
                tmpResult = objIterStart.value().at_pointer("/key")
                if tmpResult.error() == SUCCESS:
                    e.key = parseKey(tmpResult.value())
                tmpResult = objIterStart.value().at_pointer("/properties")
                if tmpResult.error() == SUCCESS:
                    innerObject = tmpResult.value().get_object()
                    objIterStartInner = innerObject.begin()
                    objIterEndInner = innerObject.end()
                    while objIterStartInner != objIterEndInner:
                        e[toPyStr(objIterStartInner.key())] = toEntityStructure(objIterStartInner.value())
                        if objIterStartInner.value().at_pointer("/arrayValue/values").error() == SUCCESS:
                            # We have to collect the non-indexed flag from the children of lists
                            arr = objIterStartInner.value().at_pointer("/arrayValue/values").value().get_array()
                            arrayIt = arr.begin()
                            arrayItEnd = arr.end()
                            allExcluded = True
                            while arrayIt != arrayItEnd:
                                element = dereference(arrayIt)
                                tmpResult = element.at_pointer("/excludeFromIndexes")
                                if tmpResult.error() != SUCCESS or not tmpResult.value().get_bool():
                                    allExcluded = False
                                    break
                                preincrement(arrayIt)
                            if allExcluded:
                                excludeList.add(toPyStr(objIterStartInner.key()))
                        else:
                            # For *all* other datatypes, we can simply check the dict it's defined in
                            tmpResult = objIterStartInner.value().at_pointer("/excludeFromIndexes")
                            if tmpResult.error() == SUCCESS and tmpResult.value().get_bool():
                                excludeList.add(toPyStr(objIterStartInner.key()))
                        preincrement(objIterStartInner)
                e.exclude_from_indexes = excludeList
                return e
            elif (strView.compare("nullValue") == 0):
                return None
            elif (strView.compare("stringValue") == 0):
                return toPyStr(objIterStart.value().get_string())
            elif (strView.compare("doubleValue") == 0):
                return objIterStart.value().get_double()
            elif (strView.compare("integerValue") == 0):
                return int(toPyStr(objIterStart.value().get_string()))
            elif (strView.compare("arrayValue") == 0):
                tmpResult = objIterStart.value().at_pointer("/values")
                if tmpResult.error() == SUCCESS:
                    return toEntityStructure(tmpResult.value())
                else:
                    return []
            elif (strView.compare("booleanValue") == 0):
                return objIterStart.value().get_bool()
            elif (strView.compare("timestampValue") == 0):
                dateStr = toPyStr(objIterStart.value().get_string())[:-1]  # Strip "Z" at the end
                if "." in dateStr:  # With milli-seconds
                    return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S.%f").replace(tzinfo=timezone.utc)
                else:
                    return datetime.strptime(dateStr, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
            elif (strView.compare("blobValue") == 0):
                return b64decode(toPyStr(objIterStart.value().get_string()))
            elif (strView.compare("keyValue") == 0):
                return parseKey(objIterStart.value())
            elif (strView.compare("geoPointValue") == 0):
                return objIterStart.value().at_key("latitude").get_double(), \
                    objIterStart.value().at_key("longitude").get_double()
            elif (strView.compare("excludeFromIndexes") == 0 or strView.compare("version") == 0):
                pass
            else:
                raise ValueError("Invalid key in entity json: %s" % toPyStr(strView))
    elif et == ARRAY:
        arr = v.get_array()
        arrayIt = arr.begin()
        arrayItEnd = arr.end()
        if isInitial:
            res = {}
        else:
            res = []
        while arrayIt != arrayItEnd:
            element = dereference(arrayIt)
            entity = toEntityStructure(element)
            if isInitial:
                res[entity.key] = entity
            else:
                res.append(entity)
            preincrement(arrayIt)
        return res

cdef inline object toPythonStructure(simdjsonElement v):
    # Convert json (sub-)tree to python objects
    cdef simdjsonElementType et = v.type()
    cdef simdjsonArray.iterator arrayIt, arrayItEnd
    cdef simdjsonObject.iterator objIterStart, objIterEnd
    cdef simdjsonElement element
    cdef simdjsonObject Object
    cdef stringView strView
    if et == OBJECT:
        Object = v.get_object()
        objIterStart = Object.begin()
        objIterEnd = Object.end()
        res = {}
        while objIterStart != objIterEnd:
            strView = objIterStart.key()
            res[toPyStr(strView)] = toPythonStructure(objIterStart.value())
            preincrement(objIterStart)
        return res
    elif et == ARRAY:
        arr = v.get_array()
        arrayIt = arr.begin()
        arrayItEnd = arr.end()
        res = []
        while arrayIt != arrayItEnd:
            element = dereference(arrayIt)
            res.append(toPythonStructure(element))
            preincrement(arrayIt)
        return res
    elif et == STRING:
        return toPyStr(v.get_string())
    elif et == INT64:
        return v.get_int64()
    elif et == UINT64:
        return v.get_uint64()
    elif et == DOUBLE:
        return v.get_double()
    elif et == NULL_VALUE:
        return None
    elif et == BOOL:
        return v.get_bool()
    else:
        raise ValueError("Unknown type")

def runSingleFilter(queryDefinition: QueryDefinition, limit: int) -> List[Entity]:
    """
        Runs a single Query as defined by queryDefinition. The limit of the queryDefinition is ignored and must
        be specified separately to prevent _calculateInternalMultiQueryLimit from modifying the queryDefinition.

        :param queryDefinition: The query to run
        :param limit:  How many entities to return at maximum
        :return: The list of entities fetched from the datastore
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element
    res = []
    internalStartCursor = None  # Will be set if we need to fetch more than one batch
    flipResults = False  # If set, we'll reverse the list returned (Sortorder was Inverted*)
    currentTxn = currentTransaction.get()
    if currentTxn:
        readOptions = {"transaction": currentTxn["key"]}
    else:
        readOptions = {"readConsistency": "STRONG"}
    while True:  # We might need to fetch more than one batch
        postData = {
            "partitionId": {
                "project_id": projectID,
            },
            "readOptions": readOptions,
            "query": {
                "kind": [
                    {
                        "name": queryDefinition.kind,
                    }
                ],
                "limit": limit - len(res)
            },
        }
        if queryDefinition.filters:
            filterList = []
            for k, v in queryDefinition.filters.items():
                key, op = k.split(" ")
                if op == "=":
                    op = "EQUAL"
                elif op == "<":
                    op = "LESS_THAN"
                elif op == "<=":
                    op = "LESS_THAN_OR_EQUAL"
                elif op == ">":
                    op = "GREATER_THAN"
                elif op == ">=":
                    op = "GREATER_THAN_OR_EQUAL"
                else:
                    raise ValueError("Invalid op %s" % op)
                if not isinstance(v, list):
                    # An entity can have a list of values for a single property, so it's possible to enforce
                    # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
                    # enforce we always have a list here
                    v = [v]
                for singleValue in v:
                    filterList.append({
                        "propertyFilter": {
                            "property": {
                                "name": key,
                            },
                            "op": op,
                            "value": pythonPropToJson(singleValue)
                        }
                    })
            if len(filterList) == 1:  # Special, single filter
                postData["query"]["filter"] = filterList[0]
            else:
                postData["query"]["filter"] = {
                    "compositeFilter": {
                        "op": "AND",
                        "filters": filterList
                    }
                }
        if queryDefinition.orders:
            postData["query"]["order"] = [
                {
                    "property": {"name": sortOrder[0]},
                    "direction": "ASCENDING" if sortOrder[1].value in [1, 4] else "DESCENDING"
                } for sortOrder in queryDefinition.orders
            ]
            flipResults = queryDefinition.orders[0][1].value > 2  # Either InvertedAscending or InvertedDescending
        if queryDefinition.distinct:
            postData["query"]["distinctOn"] = [
                {
                    "name": distinctKey
                } for distinctKey in queryDefinition.distinct
            ]
        if queryDefinition.startCursor or internalStartCursor:
            postData["query"]["startCursor"] = internalStartCursor or queryDefinition.startCursor
        if queryDefinition.endCursor:
            postData["query"]["endCursor"] = queryDefinition.endCursor
        resp = authenticated_request(
            url="https://datastore.googleapis.com/v1/projects/%s:runQuery" % projectID,
            data=json.dumps(postData).encode("UTF-8"),
        )

        is_viur_datastore_request_ok(resp)
        assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
        element = parser.parse(data_ptr, pysize, 1)
        if element.at_pointer("/batch").error() != SUCCESS:
            logging.error("INVALID RESPONSE RECEIVED")
            logging.error(json.loads(resp.content))
        #	res.update(toEntityStructure(element.at_key("batch"), isInitial=True))
        element = element.at_key("batch")
        if element.at_pointer("/entityResults").error() == SUCCESS:
            res.extend(toEntityStructure(element.at_key("entityResults"), isInitial=False))
        else:  # No results received
            if toPyStr(element.at_key("moreResults").get_string()) == "NOT_FINISHED":
                logging.warning("Query not finished. Maybe some entries are missing.")
                logging.warning("Queried %s with filter %s and orders %s. - Please create a matching index" % (
                    queryDefinition.kind, queryDefinition.filters, queryDefinition.orders))
            else:
                break
        if toPyStr(element.at_key("moreResults").get_string()) == "NO_MORE_RESULTS":
            internalStartCursor = None
        else:
            internalStartCursor = toPyStr(element.at_key("endCursor").get_string())
        if toPyStr(element.at_key("moreResults").get_string()) != "NOT_FINISHED" or len(res) == limit:
            break
    queryDefinition.currentCursor = internalStartCursor
    if conf["traceQueries"]:
        orders = queryDefinition.orders
        filters = queryDefinition.filters
        distinctOn = " distinct on %s" % str(queryDefinition.distinct) if queryDefinition.distinct else ""
        logging.debug("Queried %s with filter %s and orders %s%s. Returned %s results" % (
            queryDefinition.kind, filters, orders, distinctOn, len(res)))
    if flipResults:
        return res[::-1]
    return res

def Get(keys: Union[Key, List[Key]]) -> Union[None, Entity, List[Entity]]:
    """
        Fetches the entities determined by keys from the datastore. Returns or inserts None if a key is not found.
        :param keys: A Key or a List of Keys to fetch
        :return: The entity or None for the given key, a list of Entities/None if a list has been supplied
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element
    isMulti = True
    if isinstance(keys, Key):
        keys = [keys]
        isMulti = False
    if any(key.is_partial or key.kind is None for key in keys):
        raise InvalidArgumentError("Keys must not be partial or kind-less")
    accessLog = currentDbAccessLog.get()
    if isinstance(accessLog, set):
        accessLog.update(set(keys))

    currentTxn = currentTransaction.get()
    if currentTxn:
        readOptions = {"transaction": currentTxn["key"]}
    else:
        readOptions = {"readConsistency": "STRONG"}
    res = {}
    res_from_cache = {}
    res_from_db = {}
    untouched_keys = copy(keys)
    while keys:
        keys_for_request = keys[:300]

        if conf["memcache_client"] is not None:
            res_from_cache = cache.get(keys_for_request)
            # Convert the keys back to "class" representation
            res_from_cache = {Key.from_legacy_urlsafe(key): value
                              for key, value in res_from_cache.items()}

        missing_keys = [key for key in keys_for_request if key not in res_from_cache.keys()]
        if not missing_keys:
            # We had all keys in the memcache we can fetch the next batch now.
            keys = keys[300:]
            continue
        requested_keys = [
            {
                "partitionId": {
                    "project_id": projectID,
                },
                "path": keyToPath(x)
            }
            for x in missing_keys]
        postData = {
            "readOptions": readOptions,
            "keys": requested_keys,
        }
        resp = authenticated_request(
            url="https://datastore.googleapis.com/v1/projects/%s:lookup" % projectID,
            data=json.dumps(postData).encode("UTF-8"),
        )

        if is_viur_datastore_request_ok(resp):
            assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
            element = parser.parse(data_ptr, pysize, 1)
            if (element.at_pointer("/found").error() == SUCCESS):
                res_from_db.update(toEntityStructure(element.at_key("found"), isInitial=True))

            if (element.at_pointer("/deferred").error() == SUCCESS):
                deferred_keys = toPythonStructure(element.at_key("deferred"))
                missing_keys_paths = [keyToPath(key) for key in missing_keys]
                keys = keys[300:]
                for i, deferred_key in enumerate(deferred_keys):
                    if deferred_key in missing_keys_paths:
                        keys.insert(0, missing_keys[i])

            else:
                keys = keys[300:]
    res = res_from_db | res_from_cache
    if conf["memcache_client"] is not None:
        # Cache only the entities form db.
        cache.put({str(key): value for key, value in res_from_db.items()})

    if not isMulti:
        return res.get(untouched_keys[0])
    else:
        return [res.get(key) for key in untouched_keys]  # Sort by order of incoming keys

def Delete(keys: Union[Key, List[Key], Entity, List[Entity]]) -> None:
    """
        Deletes the entities stored under the given key(s).
        If a key is not found, it's silently ignored.
        A maximum of 300 Keys can be deleted at once.

        :param keys: A Key or a List of Keys
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element
    cdef simdjsonArray arrayElem
    if isinstance(keys, Key):
        keys = [keys]
    elif isinstance(keys, Entity):
        keys = [keys.key]
    keys = [(x.key if isinstance(x, Entity) else x) for x in keys]
    accessLog = currentDbAccessLog.get()
    if isinstance(accessLog, set):
        accessLog.update(set(keys))
    if not keys:  # We got an empty list (probably a query that returned no results), noting to do here
        return
    postData = {
        "mode": "NON_TRANSACTIONAL",  #"TRANSACTIONAL", #
        "mutations": [
            {
                "delete": pythonPropToJson(x)["keyValue"]
            }
            for x in keys
        ]
    }
    currentTxn = currentTransaction.get()
    if currentTxn:
        currentTxn["mutations"].extend(postData["mutations"])
        # Insert placeholders into affectedEntities as we receive a mutation-result for each key deleted
        currentTxn["affectedEntities"].extend([None] * len(keys))
        return
    resp = authenticated_request(
        url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
        data=json.dumps(postData).encode("UTF-8"),
    )
    if is_viur_datastore_request_ok(resp):
        assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
        element = parser.parse(data_ptr, pysize, 1)
        if (element.at_pointer("/mutationResults").error() != SUCCESS):
            logging.error(resp.content)
            raise NoMutationResultsError("No mutation results received")
        arrayElem = element.at_key("mutationResults").get_array()
        if arrayElem.size() != abs(len(keys)):
            logging.error(resp.content)
            raise ValueError("Invalid number of mutation-results received")
    if conf["memcache_client"] is not None:
        cache.delete([str(key) for key in keys])

def Put(entities: Union[Entity, List[Entity]]) -> Union[Entity, List[Entity]]:
    """
        Writes the given entities into the datastore. The entities can be from different kinds. If an entity has an
        complete key, and there's already an entity stored under the given key, it's overwritten.

        :param entities: The entities to store
        :return: The Entity or List of Entities as supplied, with partial keys replaced by full ones (unless called
            inside a transaction, in which case we return None as no Keys have been determined yet)
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element, innerArrayElem
    cdef simdjsonArray arrayElem
    cdef simdjsonArray.iterator arrayIt
    if isinstance(entities, Entity):
        entities = [entities]
    accessLog = currentDbAccessLog.get()
    if isinstance(accessLog, set):
        accessLog.update(set([x.key for x in entities if not x.key.is_partial]))
    postData = {
        "mode": "NON_TRANSACTIONAL",  # Always NON_TRANSACTIONAL; if we're inside a transaction we'll abort below
        "mutations": [
            {
                "upsert": pythonPropToJson(x)["entityValue"]
            }
            for x in entities
        ]
    }
    currentTxn = currentTransaction.get()
    if currentTxn:  # We're currently inside a transaction, just queue the changes
        currentTxn["mutations"].extend(postData["mutations"])
        currentTxn["affectedEntities"].extend(entities)
        return
    resp = authenticated_request(
        url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
        data=json.dumps(postData).encode("UTF-8"),
    )

    if is_viur_datastore_request_ok(resp):
        assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
        element = parser.parse(data_ptr, pysize, 1)
        if (element.at_pointer("/mutationResults").error() != SUCCESS):
            logging.error(resp.content)
            raise ValueError("No mutation-results received")
        arrayElem = element.at_key("mutationResults").get_array()
        if arrayElem.size() != abs(len(entities)):
            logging.error(resp.content)
            raise NoMutationResultsError("Invalid number of mutation-results received")
        arrayIt = arrayElem.begin()
        idx = 0
        while arrayIt != arrayElem.end():
            innerArrayElem = dereference(arrayIt)
            if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
                entities[idx].key = parseKey(innerArrayElem.at_key("key"))
            entities[idx].version = toPyStr(innerArrayElem.at_key("version").get_string())
            preincrement(arrayIt)
            idx += 1
        if conf["memcache_client"] is not None:
            # iter over all entities and write them to the cache
            cache.put(entities)
    return entities

def RunInTransaction(callback: callable, *args, **kwargs) -> Any:
    """
        Runs the given function inside a AID transaction.

        :param callback: The function to run inside a transaction
        :param args: Args to pass to the function
        :param kwargs: Kwargs to pass to the function
        :return: The return-value of the callback function
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element, innerArrayElem
    cdef simdjsonArray arrayElem
    cdef simdjsonArray.iterator arrayIt
    for exponential_backoff in range(1, 4):
        try:
            oldTxn = currentTransaction.get()
            allowOverriding = kwargs.pop("__allowOverriding__", None)
            if oldTxn:
                if not allowOverriding:
                    raise RecursionError("Cannot call runInTransaction while inside a transaction!")
                txnOptions = {
                    "previousTransaction": oldTxn["key"]
                }
            else:
                txnOptions = {}
            postData = {
                "transactionOptions": {
                    "readWrite": txnOptions
                }
            }
            resp = authenticated_request(
                url="https://datastore.googleapis.com/v1/projects/%s:beginTransaction" % projectID,
                data=json.dumps(postData).encode("UTF-8"),
            )
            if is_viur_datastore_request_ok(resp):
                txnKey = json.loads(resp.content)["transaction"]
                try:
                    currentTxn = {"key": txnKey, "mutations": [], "affectedEntities": []}
                    currentTransaction.set(currentTxn)
                    try:
                        res = callback(*args, **kwargs)
                    except:
                        _rollbackTxn(txnKey)
                        raise
                    if currentTxn["mutations"]:
                        # Commit TXN
                        postData = {
                            "mode": "TRANSACTIONAL",  #
                            "transaction": txnKey,
                            "mutations": currentTxn["mutations"]
                        }
                        resp = authenticated_request(
                            url="https://datastore.googleapis.com/v1/projects/%s:commit" % projectID,
                            data=json.dumps(postData).encode("UTF-8"),
                        )

                        is_viur_datastore_request_ok(resp)
                        assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
                        element = parser.parse(data_ptr, pysize, 1)
                        if (element.at_pointer("/mutationResults").error() != SUCCESS):
                            logging.error(resp.content)
                            raise NoMutationResultsError("No mutation-results received")
                        arrayElem = element.at_key("mutationResults").get_array()
                        if arrayElem.size() != abs(len(currentTxn["affectedEntities"])):
                            logging.error(resp.content)
                            raise ViurDatastoreError("Invalid number of mutation-results received")
                        arrayIt = arrayElem.begin()
                        idx = 0
                        while arrayIt != arrayElem.end():
                            innerArrayElem = dereference(arrayIt)
                            if innerArrayElem.at_pointer("/key").error() == SUCCESS:  # We got a new key assigned
                                affectedEntity = currentTxn["affectedEntities"][idx]
                                if not affectedEntity:
                                    logging.error(f"{resp.content=}")
                                    raise ViurDatastoreError("Received an unexpected key-update")
                                affectedEntity.key = parseKey(innerArrayElem.at_key("key"))
                                affectedEntity.version = toPyStr(innerArrayElem.at_key("version").get_string())
                            preincrement(arrayIt)
                            idx += 1
                        return res
                    else:  # No changes have been made - free txn
                        _rollbackTxn(txnKey)
                        return res
                finally:  # Ensure, currentTransaction is always set back to none
                    currentTransaction.set(None)
        except (CollisionError, AbortedError) as err:  # Got a collision or is aborted; retry the entire transaction
            sleep_time = 2 ** exponential_backoff
            logging.error(f"We got an error in a transaction we try again in {sleep_time} seconds")
            sleep(sleep_time)
    raise CollisionError(
        "All retries are exhausted for this transaction")  # If we made it here, all tries are exhausted

def _rollbackTxn(txnKey: str):
    """
        Internal helper that aborts the given transaction. It's important to abort pending transactions (instead
        of letting them expire server-side) so that other requests that are stalled waiting for this transaction to
        complete can continue.

        :param txnKey: The ID of the transaction that should be aborted
    """
    postData = {
        "transaction": txnKey
    }
    resp = authenticated_request(
        url="https://datastore.googleapis.com/v1/projects/%s:rollback" % projectID,
        data=json.dumps(postData).encode("UTF-8"),
    )

def AllocateIDs(keys: Union[Key, List[Key]]) -> Union[Key, List[Key]]:
    """
        Allocates numeric IDs for the keys given.
        :return: The complete Key (or a list hereof)

        .. warning: This function does not support transactions! Even if called inside transactions, the keys will
            be allocated immediately, even if the transaction aborts.
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element
    isMulti = True
    if isinstance(keys, Key):
        keys = [keys]
        isMulti = False
    keyList = [
        {
            "partitionId": {
                "project_id": projectID,
            },
            "path": keyToPath(x)
        }
        for x in keys]

    requestedKeys = keyList[:300]
    postData = {
        "keys": requestedKeys,
    }
    resp = authenticated_request(
        url="https://datastore.googleapis.com/v1/projects/%s:allocateIds" % projectID,
        data=json.dumps(postData).encode("UTF-8"),
    )
    if is_viur_datastore_request_ok(resp):
        assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
        element = parser.parse(data_ptr, pysize, 1)
        res = []
        if (element.at_pointer("/keys").error() == SUCCESS):
            arrayElem = element.at_key("keys").get_array()
            arrayIt = arrayElem.begin()
            while arrayIt != arrayElem.end():
                innerArrayElem = dereference(arrayIt)
                res.append(parseKey(innerArrayElem))
                preincrement(arrayIt)
            if not res:
                raise ValueError("Empty response received from Datastore API")
            elif not isMulti:
                return res[0]
            else:
                return res
        else:
            logging.error("Invalid data received from Datastore API")
            logging.error(resp.content)
            raise ValueError("Invalid data received from Datastore API")

def Count(kind: str = None, up_to= 2 ** 63 - 1, queryDefinition: QueryDefinition = None) -> Union[Key, List[Key]]:
    """
        Count all entries in a kind if there is only a kind is provided
        Count the entries for an given query.
        :param kind name of the module (kind) we want to count
        :param up_to can be sigend int 64 bit (max positive 2^31-1)
        :param queryDefinition: The query to run
        :return: The count as an int

        .. warning: This function does not support transactions! Even if called inside transactions, the keys will
            be allocated immediately, even if the transaction aborts.
    """
    cdef simdjsonParser parser = simdjsonParser()
    cdef Py_ssize_t pysize
    cdef char * data_ptr
    cdef simdjsonElement element
    cdef simdjsonElement element_inner
    cdef simdjsonArray array_element, array_element_inner
    logging.warning(
        "The 'Count() aggregation' query is a technical preview and cannot be guaranteed to work at this time!!!")
    if not kind:
        kind = queryDefinition.kind

    post_data = {

        "partitionId": {
            "project_id": projectID,
        }, "aggregation_query": {
            "aggregations": {"count": {"up_to": up_to}},
            "nested_query": {"kind": [{"name": kind}]}
        },

    }
    if queryDefinition:
        if queryDefinition.filters:
            filterList = []
            for k, v in queryDefinition.filters.items():
                key, op = k.split(" ")
                if op == "=":
                    op = "EQUAL"
                elif op == "<":
                    op = "LESS_THAN"
                elif op == "<=":
                    op = "LESS_THAN_OR_EQUAL"
                elif op == ">":
                    op = "GREATER_THAN"
                elif op == ">=":
                    op = "GREATER_THAN_OR_EQUAL"
                else:
                    raise ValueError("Invalid op %s" % op)
                if not isinstance(v, list):
                    # An entity can have a list of values for a single property, so it's possible to enforce
                    # more an one constraint for a a single property (e.g. x==5 and x==7 can be true), so
                    # enforce we always have a list here
                    v = [v]
                for singleValue in v:
                    filterList.append({
                        "propertyFilter": {
                            "property": {
                                "name": key,
                            },
                            "op": op,
                            "value": pythonPropToJson(singleValue)
                        }
                    })
            if len(filterList) == 1:  # Special, single filter
                post_data["aggregation_query"]["nested_query"]["filter"] = filterList[0]
            else:
                post_data["aggregation_query"]["nested_query"]["filter"] = {
                    "compositeFilter": {
                        "op": "AND",
                        "filters": filterList
                    }
                }
    resp = authenticated_request(
        url="https://datastore.googleapis.com/v1/projects/%s:runAggregationQuery" % projectID,
        data=json.dumps(post_data).encode("UTF-8"),
    )
    if is_viur_datastore_request_ok(resp):
        assert PyBytes_AsStringAndSize(resp.content, &data_ptr, &pysize) != -1
        element = parser.parse(data_ptr, pysize, 1)
        if element.at_pointer("/batch").error() != SUCCESS:
            logging.error("INVALID RESPONSE RECEIVED")
            logging.error(json.loads(resp.content))
        element = element.at_key("batch")
        # TODO  maybe this can be solved more elegant
        return int(
            toPythonStructure(element)["aggregationResults"][0]["aggregateProperties"]["property_1"]["integerValue"])



from __future__ import annotations

import logging
import typing as t
from base64 import urlsafe_b64decode, urlsafe_b64encode
from copy import deepcopy
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from viur.datastore.transport import Count, Get, runSingleFilter

from viur.datastore.config import conf
from viur.datastore.types import (
    DATASTORE_BASE_TYPES,
    Entity,
    KEY_SPECIAL_PROPERTY,
    QueryDefinition,
    SkelListRef,
    SortOrder,
    currentDbAccessLog,
)
from viur.datastore.utils import IsInTransaction

if t.TYPE_CHECKING:
    from viur.core.skeleton import SkeletonInstance


def _entryMatchesQuery(entry: Entity, singleFilter: dict) -> bool:
    """
        Utility function which checks if the given entity could have been returned by a query filtering by the
        properties in singleFilter. This can be used if a list of entities have been retrieved (e.g. by a 3rd party
        full text search engine) and these have now to be checked against the filter returned by their modules
        :meth:`viur.core.prototypes.list.listFilter` method.
        :param entry: The entity which will be tested
        :param singleFilter: A dictionary containing all the filters from the query
        :return: True if the entity could have been returned by such an query, False otherwise
    """

    def doesMatch(entryValue, requestedValue, opcode):
        if isinstance(entryValue, list):
            return any([doesMatch(x, requestedValue, opcode) for x in entryValue])
        if opcode == "=" and entryValue == requestedValue:
            return True
        elif opcode == "<" and entryValue < requestedValue:
            return True
        elif opcode == ">" and entryValue > requestedValue:
            return True
        elif opcode == "<=" and entryValue <= requestedValue:
            return True
        elif opcode == ">=" and entryValue >= requestedValue:
            return True
        return False

    for filterStr, filterValue in singleFilter.items():
        field, opcode = filterStr.split(" ")
        entryValue = entry.get(field)
        if not doesMatch(entryValue, filterValue, opcode):
            return False
    return True


class Query(object):
    """
        Base Class for querying the datastore. It's API is similar to the google.cloud.datastore.query API,
        but it provides the necessary hooks for relational or random queries, the fulltext search as well as support
        for IN filters.
    """

    def __init__(self, kind: str, srcSkelClass: Union["SkeletonInstance", None] = None, *args, **kwargs):
        """
            Constructs a new Query.
            :param kind: The kind to run this query on. This may be later overridden to run on a different kind (like
                viur-relations), but it's guaranteed to return only entities of that kind.
            :param srcSkelClass: If set, enables data-model depended queries (like relational queries) as well as the
                :meth:fetch method
        """
        super(Query, self).__init__()
        self.kind = kind
        self.srcSkel = srcSkelClass
        self.queries: Union[None, QueryDefinition, List[QueryDefinition]] = QueryDefinition(kind, {}, [])
        cbSignature = Union[None, Callable[[Query, str, Union[DATASTORE_BASE_TYPES, List[DATASTORE_BASE_TYPES]]], Union[
            None, Tuple[str, Union[DATASTORE_BASE_TYPES, List[DATASTORE_BASE_TYPES]]]]]]
        self._filterHook: cbSignature = None
        self._orderHook: cbSignature = None
        # Sometimes, the default merge functionality from MultiQuery is not sufficient
        self._customMultiQueryMerge: Union[None, Callable[[Query, List[List[Entity]], int], List[Entity]]] = None
        # Some (Multi-)Queries need a different amount of results per subQuery than actually returned
        self._calculateInternalMultiQueryLimit: Union[None, Callable[[Query, int], int]] = None
        # Allow carrying custom data along with the query. Currently only used by spartialBone to record the guranteed correctnes
        self.customQueryInfo = {}
        self.origKind = kind
        self._lastEntry = None
        self._fulltextQueryString: Union[None, str] = None
        self.lastCursor = None
        if not kind.startswith("viur") and not kwargs.get("_excludeFromAccessLog"):
            accessLog = currentDbAccessLog.get()
            if isinstance(accessLog, set):
                accessLog.add(kind)

    def setFilterHook(self, hook: Callable) -> Optional[Callable]:
        """
            Installs *hook* as a callback function for new filters.

            *hook* will be called each time a new filter constrain is added to the query.
            This allows e. g. the relationalBone to rewrite constrains added after the initial
            processing of the query has been done (e. g. by ``listFilter()`` methods).

            :param hook: The function to register as callback.
                A value of None removes the currently active hook.
            :returns: The previously registered hook (if any), or None.
        """
        old = self._filterHook
        self._filterHook = hook
        return old

    def setOrderHook(self, hook: Callable) -> Callable:
        """
            Installs *hook* as a callback function for new orderings.

            *hook* will be called each time a :func:`db.Query.order` is called on this query.

            :param hook: The function to register as callback.
                A value of None removes the currently active hook.
            :returns: The previously registered hook (if any), or None.
        """
        old = self._orderHook
        self._orderHook = hook
        return old

    def mergeExternalFilter(self, filters: dict) -> 'Query':
        """
            Safely merges filters according to the data model.

            Its only valid to call this function if the query has been created using
            :func:`server.skeleton.Skeleton.all`.

            Its safe to pass filters received from an external source (a user);
            unknown/invalid filters will be ignored, so the query-object is kept in a
            valid state even when processing malformed data.

            If complex queries are needed (e.g. filter by relations), this function
            shall also be used.

            See also :func:`server.db.Query.filter` for simple filters.

            :param filters: A dictionary of attributes and filter pairs.
            :returns: Returns the query itself for chaining.
        """
        if self.srcSkel is None:
            raise NotImplementedError("This query has not been created using skel.all()")
        if self.queries is None:  # This query is allready unsatifiable and adding more constrains to this wont change this
            return self
        skel = self.srcSkel
        if "search" in filters:
            if self.srcSkel.customDatabaseAdapter and self.srcSkel.customDatabaseAdapter.providesFulltextSearch:
                self._fulltextQueryString = str(filters["search"])
            else:
                logging.warning(
                    "Got a fulltext search query for %s which does not have a suitable customDatabaseAdapter"
                    % self.srcSkel.kindName
                )
                self.queries = None
        bones = [(y, x) for x, y in skel.items()]
        try:
            # Process filters first
            for bone, key in bones:
                bone.buildDBFilter(key, skel, self, filters)
            # Parse orders
            for bone, key in bones:
                bone.buildDBSort(key, skel, self, filters)
        except RuntimeError as e:
            logging.exception(e)
            self.queries = None
            return self
        startCursor = endCursor = None
        if "cursor" in filters and filters["cursor"] and filters["cursor"].lower() != "none":
            startCursor = filters["cursor"]
        if "endcursor" in filters and filters["endcursor"] and filters["endcursor"].lower() != "none":
            endCursor = filters["endcursor"]
        if startCursor or endCursor:
            self.setCursor(startCursor, endCursor)
        if "limit" in filters and str(filters["limit"]).isdigit() and int(filters["limit"]) > 0 and int(
            filters["limit"]) <= 100:
            self.limit(int(filters["limit"]))
        return self

    def filter(self, prop: str, value: Union[DATASTORE_BASE_TYPES, List[DATASTORE_BASE_TYPES]]) -> 'Query':
        """
            Adds a new constraint to this query.

            The following examples are equivalent: ``filter( "name", "John" )``
            and ``filter( {"name": "John"} )``.

            See also :func:`server.db.Query.mergeExternalFilter` for a safer filter implementation.

            :param prop: Name of the property + operation we'll filter by
            :param value: The value of that filter. Only valid, if *key* is a string.
            :returns: Returns the query itself for chaining.
        """
        if self.queries is None:
            # This query is already unsatisfiable and adding more constrains to this won't change this
            return self
        if self._filterHook is not None:
            try:
                r = self._filterHook(self, prop, value)
            except RuntimeError:
                self.queries = None
                return self
            if r is None:
                # The Hook did something special directly on 'self' to apply that filter,
                # no need for us to do anything
                return self
            prop, value = r
        if " " not in prop:
            # Ensure that an equality filter is explicitly postfixed with " ="
            field = prop
            op = "="
        else:
            field, op = prop.split(" ")
        if op.lower() in {"!=", "in"}:
            if isinstance(self.queries, list):
                raise NotImplementedError("You cannot use multiple IN or != filter")
            origQuery = self.queries
            self.queries = []
            if op == "!=":
                newFilter = deepcopy(origQuery)
                newFilter.filters["%s <" % field] = value
                self.queries.append(newFilter)
                newFilter = deepcopy(origQuery)
                newFilter.filters["%s >" % field] = value
                self.queries.append(newFilter)
            else:  # IN filter
                if not (isinstance(value, list) or isinstance(value, tuple)):
                    raise ValueError("Value must be list or tuple if using IN filter!")
                for val in value:
                    newFilter = deepcopy(origQuery)
                    newFilter.filters["%s =" % field] = val
                    self.queries.append(newFilter)
        else:
            filterStr = "%s %s" % (field, op)
            if isinstance(self.queries, list):
                for singeFilter in self.queries:
                    if filterStr not in singeFilter.filters:
                        singeFilter.filters[filterStr] = value
                    else:
                        if not isinstance(singeFilter.filters[filterStr]):
                            singeFilter.filters[filterStr] = [singeFilter.filters[filterStr]]
                        singeFilter.filters[filterStr].append(value)
            else:  # It must be still a dict (we tested for None already above)
                if filterStr not in self.queries.filters:
                    self.queries.filters[filterStr] = value
                else:
                    if not isinstance(self.queries.filters[filterStr], list):
                        self.queries.filters[filterStr] = [self.queries.filters[filterStr]]
                    self.queries.filters[filterStr].append(value)
            if op in {"<", "<=", ">", ">="}:
                if isinstance(self.queries, list):
                    for queryObj in self.queries:
                        if not queryObj.orders or queryObj.orders[0][0] != field:
                            queryObj.orders = [(field, SortOrder.Ascending)] + (queryObj.orders or [])
                else:
                    if not self.queries.orders or self.queries.orders[0][0] != field:
                        self.queries.orders = [(field, SortOrder.Ascending)] + (self.queries.orders or [])
        return self

    def order(self, *orderings: Tuple[str, 'SortOrder']) -> 'Query':
        """
            Specify a query sorting.

            Resulting entities will be sorted by the first property argument, then by the
            second, and so on.

            The following example

            .. code-block:: python

                query = Query( "Person" )
                query.order(("bday" db.SortOrder.Ascending), ("age", db.SortOrder.Descending))

            sorts every Person in order of their birthday, starting with January 1.
            People with the same birthday are sorted by age, oldest to youngest.


            ``order()`` may be called multiple times. Each call resets the sort order
            from scratch.

            If an inequality filter exists in this Query it must be the first property
            passed to ``order()``. Any number of sort orders may be used after the
            inequality filter property. Without inequality filters, any number of
            filters with different orders may be specified.

            Entities with multiple values for an order property are sorted by their
            lowest value.

            Note that a sort order implies an existence filter! In other words,
            Entities without the sort order property are filtered out, and *not*
            included in the query results.

            If the sort order property has different types in different entities -
            e.g. if bob['id'] is an int and fred['id'] is a string - the entities will be
            grouped first by the property type, then sorted within type. No attempt is
            made to compare property values across types.


            :param orderings: The properties to sort by, in sort order.
                Each argument must be a (string, direction) 2-tuple.
            :returns: Returns the query itself for chaining.
        """
        if self.queries is None:
            # This Query is unsatisfiable - don't try to bother
            return self

        # Check for correct order subscript
        orders = []
        for order in orderings:
            if isinstance(order, str):
                order = (order, SortOrder.Ascending)

            if not (isinstance(order[0], str) and isinstance(order[1], SortOrder)):
                raise TypeError(
                    f"Invalid ordering {order}, it has to be a tuple. Try: `(\"{order}\", SortOrder.Ascending)`")

            orders.append(order)

        orderings = tuple(orders)

        if self._orderHook is not None:
            try:
                orderings = self._orderHook(self, orderings)
            except RuntimeError:
                self.queries = None
                return self
            if orderings is None:
                return self

        if isinstance(self.queries, list):
            for query in self.queries:
                query.orders = list(orderings)
        else:
            self.queries.orders = list(orderings)

        return self

    def setCursor(self, startCursor: str, endCursor: Optional[str] = None) -> 'Query':
        """
            Sets the start and optionally end cursor for this query.

            The result set will only include results between these cursors.
            The cursor is generated by an earlier query with exactly the same configuration.

            Its safe to use client-supplied cursors, a cursor can't be abused to access entities
            which don't match the current filters.

            :param startCursor: The start cursor for this query.
            :param endCursor: The end cursor for this query.
            :returns: Returns the query itself for chaining.
        """
        if isinstance(self.queries, list):
            for query in self.queries:
                assert isinstance(query, QueryDefinition)
                if startCursor:
                    query.startCursor = urlsafe_b64decode(startCursor.encode("ASCII")).decode("ASCII")
                if endCursor:
                    query.endCursor = urlsafe_b64decode(endCursor.encode("ASCII")).decode("ASCII")
        else:
            assert isinstance(self.queries, QueryDefinition)
            if startCursor:
                self.queries.startCursor = urlsafe_b64decode(startCursor.encode("ASCII")).decode("ASCII")
            if endCursor:
                self.queries.endCursor = urlsafe_b64decode(endCursor.encode("ASCII")).decode("ASCII")
        return self

    def limit(self, limit: int) -> 'Query':
        """
            Sets the query limit to *amount* entities in the result.

            Specifying an limit of 0 disables the limit (use with care!).

            :param limit: The maximum number of entities.
            :returns: Returns the query itself for chaining.
        """
        if isinstance(self.queries, QueryDefinition):
            self.queries.limit = limit
        elif isinstance(self.queries, list):
            for query in self.queries:
                query.limit = limit
        return self

    def distinctOn(self, keyList: List[str]) -> 'Query':
        """
            Ensure only entities with distinct values on the fields listed are returned.
            This will implicitly override your SortOrder as all fields listed in keyList have to be sorted first.
        """
        if isinstance(self.queries, QueryDefinition):
            self.queries.distinct = keyList
        elif isinstance(self.queries, list):
            for query in self.queries:
                query.distinct = keyList
        return self

    def getCursor(self) -> Optional[str]:
        """
            Get a valid cursor from the last run of this query.

            The source of this cursor varies depending on what the last call was:
            - :func:`server.db.Query.run`: A cursor that points immediately behind the
                last result pulled off the returned iterator.
            - :func:`server.db.Query.get`:: A cursor that points immediately behind the
                last result in the returned list.

            :returns: A cursor that can be used in subsequent query requests or None if that query does not support
                cursors or ther're no more elements to fetch
        """
        if isinstance(self.queries, QueryDefinition):
            q = self.queries
        elif isinstance(self.queries, list):
            q = self.queries[0]
        return urlsafe_b64encode(q.currentCursor.encode("ASCII")).decode("ASCII") if q.currentCursor else None

    def get_orders(self) -> List[Tuple[str, SortOrder]] | None:
        """
            Get the orders from this query.

            :returns: The orders form this query as a list if there is no orders set it returns None
        """
        q = self.queries

        if isinstance(q, (list, tuple)):
            q = q[0]

        if not isinstance(q, QueryDefinition):
            raise ValueError(
                f"self.queries can only be a 'QueryDefinition' or a list of, but found {self.queries!r}"
            )

        return q.orders or None

    def getKind(self) -> str:
        """
            :returns: the *current* kind of this query. This may not be the kind this query has been constructed with
                as relational bones may rewrite this.
        """
        return self.kind

    def _runSingleFilterQuery(self, query: QueryDefinition, limit: int) -> List[Entity]:
        """
            Internal helper function that runs a single query definition on the datastore and returns a list of
            entities found.
            :param query: The querydefinition (filters, orders, distinct etc) to run against the datastore
            :param limit: How many results shoult at most be returned
            :return: The first *limit* entities that matches this query
        """
        return runSingleFilter(query, limit)

    def _mergeMultiQueryResults(self, inputRes: List[List[Entity]]) -> List[Entity]:
        """
            Merge the lists of entries into a single list; removing duplicates and restoring sort-order
            :param inputRes: Nested Lists of Entries returned by each individual query run
            :return: Sorted & deduplicated list of entries
        """
        seenKeys = set()
        res = []
        for subList in inputRes:
            for entry in subList:
                key = entry.key
                if key in seenKeys:
                    continue
                seenKeys.add(key)
                res.append(entry)
        # Fixme: What about filters that mix different inequality filters - we'll now simply ignore any implicit sortorder
        return self._resortResult(res, {}, self.queries[0].orders)

    def _resortResult(self, entities: List[Entity], filters: Dict[str, DATASTORE_BASE_TYPES],
                      orders: List[Tuple[str, 'SortOrder']]) -> List[Entity]:
        """
            Internal helper that takes a (deduplicated) list of entities that has been fetched from different internal
            queries (the datastore does not support IN filters itself, so we have to query each item in that array
            separately) and resorts the list so it matches the query again.

            :param entities: List of entities to resort
            :param filters: The filter used in the query (used to determine implicit sort order by an inequality filter)
            :param orders: The sort-orders to apply
            :return: The sorted list
        """

        def getVal(src: Entity, fieldVars: Union[str, Tuple[str]], direction: SortOrder) -> Any:
            # Descent into the target until we reach the property we're looking for
            if isinstance(fieldVars, tuple):
                for fv in fieldVars:
                    if not fv in src:
                        return None
                    src = src[fv]
            else:
                if not fieldVars in src:
                    return (str(type(None)), 0)
                src = src[fieldVars]
            # Lists are handled differently, here the smallest or largest value determines it's position in the result
            if isinstance(src, list) and len(src):
                try:
                    src.sort()
                except TypeError:
                    # It's a list of dicts or the like for which no useful sort-order is specified
                    pass
                if direction == SortOrder.Ascending:
                    src = src[0]
                else:
                    src = src[-1]
            # We must return this tuple because inter-type comparison isn't possible in Python3 anymore
            return str(type(src)), src if src is not None else 0

        # Check if we have an inequality filter which implies an sortorder
        ineqFilter = None
        for k, _ in filters.items():
            end = k[-2:]
            if "<" in end or ">" in end:
                ineqFilter = k.split(" ")[0]
                break
        if ineqFilter and (not orders or not orders[0][0] == ineqFilter):
            orders = [(ineqFilter, SortOrder.Ascending)] + (orders or [])

        for orderField, direction in orders[::-1]:
            if orderField == KEY_SPECIAL_PROPERTY:
                pass  # FIXME !!
            # entities.sort(key=lambda x: x.key, reverse=direction == SortOrder.Descending)
            else:
                try:
                    entities.sort(key=partial(getVal, fieldVars=orderField, direction=direction),
                                  reverse=direction == SortOrder.Descending)
                except TypeError:
                    # We hit some incomparable types
                    pass
        return entities

    def _fixKind(self, resultList: List[Entity]) -> List[Entity]:
        """
            Jump to parentKind if necessary (used in relations)
        """
        resultList = list(resultList)
        if resultList and resultList[0].key.kind != self.origKind and resultList[0].key.parent and \
            resultList[0].key.parent.kind == self.origKind:
            return list(Get(list(dict.fromkeys([x.key.parent for x in resultList]))))
        return resultList

    def run(self, limit: int = -1) -> List[Entity]:
        """
            Run this query.

            It is more efficient to use *limit* if the number of results is known.

            If queried data is wanted as instances of Skeletons, :func:`server.db.Query.fetch`
            should be used.

            :param limit: Limits the query to the defined maximum entities.

            :returns: The list of found entities

            :raises: :exc:`BadFilterError` if a filter string is invalid
            :raises: :exc:`BadValueError` if a filter value is invalid.
            :raises: :exc:`BadQueryError` if an IN filter in combination with a sort order on\
            another property is provided
        """
        if self.queries is None:
            if conf["traceQueries"]:
                logging.debug("Query on %s aborted as being not satisfiable" % self.kind)
            return []

        if self._fulltextQueryString:
            if IsInTransaction():
                raise ValueError("Can't run fulltextSearch inside transactions!")  # InvalidStateError FIXME!
            qryStr = self._fulltextQueryString
            self._fulltextQueryString = None  # Reset, so the adapter can still work with this query
            res = self.srcSkel.customDatabaseAdapter.fulltextSearch(qryStr, self)
            if not self.srcSkel.customDatabaseAdapter.fulltextSearchGuaranteesQueryConstrains:
                # Search might yield results that are not included in the listfilter
                if isinstance(self.queries, QueryDefinition):  # Just one
                    res = [x for x in res if _entryMatchesQuery(x, self.queries.filters)]
                else:  # Multi-Query, must match at least one
                    res = [x for x in res if any([_entryMatchesQuery(x, y.filters) for y in self.queries])]
        elif isinstance(self.queries, list):
            # We have more than one query to run
            if self._calculateInternalMultiQueryLimit:
                limit = self._calculateInternalMultiQueryLimit(self, limit if limit != -1 else self.queries[0].limit)
            res = []
            # We run all queries first (preventing multiple round-trips to the server)
            for singleQuery in self.queries:
                res.append(self._runSingleFilterQuery(singleQuery, limit if limit != -1 else singleQuery.limit))
            # Wait for the actual results to arrive and convert the protobuffs to Entries
            res = [self._fixKind(x) for x in res]
            if self._customMultiQueryMerge:
                # We have a custom merge function, use that
                res = self._customMultiQueryMerge(self, res, limit if limit != -1 else self.queries[0].limit)
            else:
                # We must merge (and sort) the results ourself
                res = self._mergeMultiQueryResults(res)
        else:  # We have just one single query
            res = self._fixKind(self._runSingleFilterQuery(self.queries, limit if limit != -1 else self.queries.limit))
        if res:
            self._lastEntry = res[-1]
        return res

    def count(self, up_to: int = 2 ** 63 - 1) -> int:
        """
            The count operation cost one entity read for up to 1,000 index entries matched
            (https://cloud.google.com/datastore/docs/aggregation-queries#pricing)
            :param up_to can be sigend int 64 bit (max positive 2^31-1)

            :returns: Count entries for this query.
        """
        if self.queries is None:
            if conf["traceQueries"]:
                logging.debug("Query on %s aborted as being not satisfiable" % self.kind)
            return -1
        elif isinstance(self.queries, list):
            raise ValueError("No count on Multiqueries")
        else:
            return Count(queryDefinition=self.queries, up_to=up_to)

    def fetch(self, limit: int = -1) -> SkelListRef['SkeletonInstance'] | None:
        """
            Run this query and fetch results as :class:`server.skeleton.SkelList`.

            This function is similar to :func:`server.db.Query.run`, but returns a
            :class:`server.skeleton.SkelList` instance instead of Entities.

            :warning: The query must be limited!

            If queried data is wanted as instances of Entity, :func:`server.db.Query.run`
            should be used.

            :param limit: Limits the query to the defined maximum entities.
                A maxiumum value of 99 entries can be fetched at once.

            :raises: :exc:`BadFilterError` if a filter string is invalid
            :raises: :exc:`BadValueError` if a filter value is invalid.
            :raises: :exc:`BadQueryError` if an IN filter in combination with a sort order on
                another property is provided
        """
        assert conf["SkeletonInstanceRef"] is not None, "conf['SkeletonInstanceRef'] has not been set!"
        if self.srcSkel is None:
            raise NotImplementedError("This query has not been created using skel.all()")
        # limit = limit if limit != -1 else self._limit
        if limit != -1 and not (0 < limit <= 100):
            logging.error(("Limit", limit))
            raise NotImplementedError(
                "This query is not limited! You must specify an upper bound using limit() between 1 and 100")
        dbRes = self.run(limit)
        if dbRes is None:
            return None
        res = SkelListRef(self.srcSkel)
        for e in dbRes:
            skelInstance = conf["SkeletonInstanceRef"](self.srcSkel.skeletonCls, clonedBoneMap=self.srcSkel.boneMap)
            skelInstance.dbEntity = e
            res.append(skelInstance)
        res.getCursor = lambda: self.getCursor()
        res.get_orders = lambda: self.get_orders()
        return res

    def iter(self) -> t.Iterator[Entity]:
        """
            Run this query and return an iterator for the results.

            The advantage of this function is, that it allows for iterating
            over a large result-set, as it hasn't have to be pulled in advance
            from the data store.

            This function intentionally ignores a limit set by :func:`server.db.Query.limit`.

            :warning: If iterating over a large result set, make sure the query supports cursors. \
            Otherwise, it might not return all results as the AppEngine doesn't maintain the view \
            for a query for more than ~30 seconds.
        """
        if self.queries is None:  # Noting to pull here
            raise StopIteration()
        elif isinstance(self.queries, list):
            raise ValueError("No iter on Multiqueries")
        while True:
            qryRes = self._runSingleFilterQuery(self.queries, 20)
            yield from qryRes
            if not self.queries.currentCursor:  # We reached the end of that query
                break
            self.queries.startCursor = self.queries.currentCursor

    def getEntry(self) -> Union[None, Entity]:
        """
            Returns only the first entity of the current query.

            :returns: The first entity on success, or None if the result-set is empty.
        """
        try:
            res = list(self.run(limit=1))[0]
            return res
        except (IndexError, TypeError):  # Empty result-set
            return None

    def getSkel(self) -> Optional['SkeletonInstance']:
        """
            Returns a matching :class:`server.db.skeleton.Skeleton` instance for the
            current query.

            Its only possible to use this function if this query has been created using
            :func:`server.skeleton.Skeleton.all`.

            :returns: The Skeleton or None if the result-set is empty.
        """
        if self.srcSkel is None:
            raise NotImplementedError("This query has not been created using skel.all()")
        res = self.getEntry()
        if res is None:
            return None
        self.srcSkel.setEntity(res)
        return self.srcSkel

    def clone(self) -> 'Query':
        """
            Returns a deep copy of the current query.

            :returns: The cloned query.
        """
        res = Query(self.getKind(), self.srcSkel)
        res.kind = self.kind
        res.queries = deepcopy(self.queries)
        # res.filters = deepcopy(self.filters)
        # res.orders = deepcopy(self.orders)
        # res._limit = self._limit
        res._filterHook = self._filterHook
        res._orderHook = self._orderHook
        # res._startCursor = self._startCursor
        # res._endCursor = self._endCursor
        res._customMultiQueryMerge = self._customMultiQueryMerge
        res._calculateInternalMultiQueryLimit = self._calculateInternalMultiQueryLimit
        res.customQueryInfo = self.customQueryInfo
        res.origKind = self.origKind
        res._fulltextQueryString = self._fulltextQueryString
        # res._distinct = self._distinct
        return res

    def __repr__(self):
        return "<db.Query on %s with queries %s>" % (self.kind, self.queries)



import sys
import time
import time as time_module
from typing import Any, Dict, List, Union

import logging
from viur.datastore.config import conf
from viur.datastore.types import Entity, Key

MEMCACHE_MAX_BATCH_SIZE = 30
MEMCACHE_NAMESPACE = "viur-datastore"
MEMCACHE_TIMEOUT = 60 * 60
MEMCACHE_MAX_SIZE = 1_000_000

"""

	This Module controls the Interaction with the Memcache from Google
	To activate the cache copy this code in your main.py
	..  code-block:: python
	# Example
	from viur.core import db
	if not conf["viur.instance.is_dev_server"]:
		from google.appengine.api.memcache import Client
		db.config["memcache_client"] = Client()
	else:
		db.config["memcache_client"] = db.cache.LocalMemcache()

"""

__all__ = [
    "MEMCACHE_MAX_BATCH_SIZE",
    "MEMCACHE_NAMESPACE",
    "MEMCACHE_TIMEOUT",
    "MEMCACHE_MAX_SIZE",
    "get",
    "put",
    "delete",
    "LocalMemcache"
]


def get(keys: Union[str, Key, List[str], List[Key]]) -> Dict[str, dict]:
    """
        Reads data form the memcache.
        :param Union[str, Key, List[str], List[Key]] keys: Unique identifier(s) for one or more entry(s).
        :return: A dict with the entry(s) that found in the memcache.
    """
    if not check_for_memcache():
        return {}
    if not isinstance(keys, list):
        keys = [keys]
    keys = [str(key) for key in keys]  # Enforce that all keys are strings
    res = {}
    try:
        while keys:
            res |= conf["memcache_client"].get_multi(keys[:MEMCACHE_MAX_BATCH_SIZE], namespace=MEMCACHE_NAMESPACE)
            keys = keys[MEMCACHE_MAX_BATCH_SIZE:]
    except Exception as e:
        logging.error(f"""Failed to get keys form the memcache with {e=}""")
    return res


def put(data: Union[Entity, Dict[Key, Entity], List[Entity]]):
    """
        Writes Data to the memcache.

        :param Union[Entity, Dict[Key, Entity], List[Entity]] data: Data to write
    """
    if not check_for_memcache():
        return
    if isinstance(data, list):
        data = {item.key: item for item in data}
    elif isinstance(data, Entity):
        data = {data.key: data}
    elif not isinstance(data, dict):
        raise TypeError(f"Invalid type {type(data)}. Expected a db.Entity, list or dict.")
    # Add only values to cache <= MEMMAX_SIZE (1.000.000)
    data = {str(key): value for key, value in data.items() if get_size(value) <= MEMCACHE_MAX_SIZE}

    keys = list(data.keys())
    try:
        while keys:
            data_batch = {key: data[key] for key in keys[:MEMCACHE_MAX_BATCH_SIZE]}
            conf["memcache_client"].set_multi(data_batch, namespace=MEMCACHE_NAMESPACE, time=MEMCACHE_TIMEOUT)
            keys = keys[MEMCACHE_MAX_BATCH_SIZE:]
    except Exception as e:
        logging.error(f"""Failed to put data to the memcache with {e=}""")


def delete(keys: Union[str, Key, List[str], List[Key]]) -> None:
    """
        Deletes an Entry form memcache.

        :param Union[str, Key, List[str], List[Key]] keys: Unique identifier(s) for one or more entry(s).
    """
    if not check_for_memcache():
        return
    if not isinstance(keys, list):
        keys = [keys]
    keys = [str(key) for key in keys]  # Enforce that all keys are strings
    try:
        while keys:
            conf["memcache_client"].delete_multi(keys[:MEMCACHE_MAX_BATCH_SIZE], namespace=MEMCACHE_NAMESPACE)
            keys = keys[MEMCACHE_MAX_BATCH_SIZE:]
    except Exception as e:
        logging.error(f"""Failed to delete keys form the memcache with {e=}""")


def flush():
    """
        Deletes everything in memcache.
    """
    if not check_for_memcache():
        return
    try:
        conf["memcache_client"].flush_all()
    except Exception as e:
        logging.error(f"""Failed to flush the memcache with {e=}""")


def get_size(obj: Any) -> int:
    """
        Utility function that counts the size of an object in bytes.
    """
    if isinstance(obj, dict):
        return sum(get_size([k, v]) for k, v in obj.items())
    elif isinstance(obj, list):
        return sum(get_size(x) for x in obj)

    return sys.getsizeof(obj)


def check_for_memcache() -> bool:
    if conf["memcache_client"] is None:
        logging.warning(f"""conf["memcache_client"] is 'None'. It can not be used.""")
        return False
    return True


class LocalMemcache:
    def __init__(self):
        self._data = {}

    def get_multi(self, keys: List[str], namespace: str = MEMCACHE_NAMESPACE):
        self._data.setdefault(namespace, {})
        res = {}
        for key in keys:
            if (data := self._data[namespace].get(key)) is not None:
                if data["__lifetime__"]["last_seen"] + data["__lifetime__"]["timeout"] > time.time():
                    res[key] = data["__data__"]
                else:
                    self._data[namespace].pop(key)
        return res

    def set_multi(self, data: Dict[str, Any], namespace: str = MEMCACHE_NAMESPACE, time: int = MEMCACHE_TIMEOUT):
        self._data.setdefault(namespace, {})
        for key, value in data.items():
            self._data[namespace][key] = {}
            self._data[namespace][key]["__data__"] = value
            self._data[namespace][key]["__lifetime__"] = {"timeout": time, "last_seen": time_module.time()}

    def delete_multi(self, keys: List[str] = [], namespace: str = MEMCACHE_NAMESPACE):
        self._data.setdefault(namespace, {})
        for key in keys:
            if (data := self._data[namespace].get(key)) is not None:
                self._data[namespace].pop(key)

    def flush_all(self):
        self._data.clear()



"""
	This file just holds some configuration variables that will influence
	the behaviour of this library.
"""
conf = {
    # If set, we'll log each query we run
    "traceQueries": False,

    # A reference to the skeleton container of ViUR. Unless set, fetch() and getSkel() will fail
    "SkeletonInstanceRef": None,
    # an allow list which error codes should trigger a verbose message output on stderr/stdout
    # take a look in viur.datastore.errors.CANONICAL_ERROR_CODE_MAP keys and the Exceptions for reference
    "verbose_error_codes": {
        "ABORTED",
        "ALREADY_EXISTS",
        "DEADLINE_EXCEEDED",
        "FAILED_PRECONDITION",
        "INTERNAL",
        "INVALID_ARGUMENT",
        "NOT_FOUND",
        "PERMISSION_DENIED",
        "RESOURCE_EXHAUSTED",
        "UNAUTHENTICATED",
        "UNAVAILABLE"
    },
    # A Client form the Google Memcache Library.
    "memcache_client": None,
}



from setuptools import setup, Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext

setup(
    name='viur-datastore',
    version="1.3.14rc1",
    author="Tobias Steinrücken, Stefan Kögl",
    author_email="devs@viur.dev",
    maintainer="Stefan Kögl",
    maintainer_email="devs@viur.dev",
    description="A faster replacement for google-cloud-datastore",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/viur-framework/viur-datastore",
    packages=['viur.datastore'],
    package_dir={'': 'src'},
    python_requires=">=3.10",
    install_requires=[
        "google-cloud-datastore>=2.0",
        "google-auth>=2.0",
        "requests>=2.0",
    ],
    cmdclass={'build_ext': build_ext},
    ext_modules=cythonize([Extension("viur.datastore.transport", ["src/viur/datastore/transport.pyx"], language="c++",
                                     extra_compile_args=["-std=c++11"])]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Database :: Front-Ends",
    ],
)
